;; Generated package description from paced.el  -*- no-byte-compile: t -*-
(define-package "paced" "1.1.3.0.2019022814125" "Predictive Abbreviation Completion and Expansion using Dictionaries" '((emacs "25.1") (async "1.9.1")) :keywords '("convenience" "completion") :authors '(("Ian Dunn" . "dunni@gnu.org")) :maintainer '("Ian Dunn" . "dunni@gnu.org") :url "https://savannah.nongnu.org/projects/paced-el/")
